library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

package pkg_exp is
  component exp_11 is
    port (x : in  std_logic_vector(11 - 1 downto 0);
          y : out std_logic_vector(12 - 1 downto 0));
  end component;
  component exp_15 is
    port (x : in  std_logic_vector(15 - 1 downto 0);
          y : out std_logic_vector(16 - 1 downto 0));
  end component;
  component exp_19 is
    port (x : in  std_logic_vector(19 - 1 downto 0);
          y : out std_logic_vector(20 - 1 downto 0));
  end component;
  component exp_21 is
    port (x : in  std_logic_vector(21 - 1 downto 0);
          y : out std_logic_vector(22 - 1 downto 0);
          sign : in std_logic);
  end component;
end package;

-- exp d'un nombre avec une mantisse de 22 bits dont 11 non nuls
-- =============================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
library work;
use work.pkg_exp_tbl.all;

entity exp_11 is
  port (x : in  std_logic_vector(11 - 1 downto 0);
        y : out std_logic_vector(12 - 1 downto 0));
end entity;

architecture arch of exp_11 is
  signal part_1 : std_logic_vector(11 - 1 downto 0);
  -- exponentielle du morceau moins 1
  signal exp_part1 : std_logic_vector(11 - 1 downto 0);
begin
  -- premiere partie
  part_1 <= x;
  exp_part1 <= part_1;
  y <= ("0" & exp_part1);
end architecture;

-- exp d'un nombre avec une mantisse de 22 bits dont 15 non nuls
-- =============================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
library work;
use work.pkg_exp_tbl.all;
use work.pkg_exp.exp_11;

entity exp_15 is
  port (x : in  std_logic_vector(15 - 1 downto 0);
        y : out std_logic_vector(16 - 1 downto 0));
end entity;

architecture arch of exp_15 is
  signal part_1 : std_logic_vector(15 - 1 downto 11);
  -- valeur lue dans la table pour ce morceau
  signal tbl_out : std_logic_vector(7 - 1 downto 0);
  -- exponentielle du morceau moins 1
  signal exp_part1 : std_logic_vector(15 - 1 downto 0);
  -- exponentielle des morceaux suivants moins 1
  signal exp_part2 : std_logic_vector(12 - 1 downto 0);
  -- produit des deux signaux precedents
  signal product : std_logic_vector(11 - 1 downto 0);
begin
  -- premiere partie
  part_1 <= x(15 - 1 downto 11);
  -- exponentielle de cette partie
  component1 : exp_tbl_15
    port map (x => part_1,
              y => tbl_out);
  exp_part1 <= part_1 & ("0000" & tbl_out);

  -- exponentielle de la partie restante
  component2 : exp_11
    port map (x => x(11 - 1 downto 0),
              y => exp_part2);

  -- calcul du resultat
  product <= exp_part1(15 - 1 downto 10) * exp_part2(12 - 1 downto 6);

  y <= ("0" & exp_part1) + ("0000" & exp_part2) + ("00000000000" & product(11 - 1 downto 6));
end architecture;

-- exp d'un nombre avec une mantisse de 22 bits dont 19 non nuls
-- =============================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
library work;
use work.pkg_exp_tbl.all;
use work.pkg_exp.exp_15;

entity exp_19 is
  port (x : in  std_logic_vector(19 - 1 downto 0);
        y : out std_logic_vector(20 - 1 downto 0));
end entity;

architecture arch of exp_19 is
  signal exp_tbl_out : std_logic_vector(15 - 1 downto 14);
  signal part_1 : std_logic_vector(19 - 1 downto 14);
  signal exp_part1 : std_logic_vector(6 - 1 downto 0);
  signal remainder_1 : std_logic_vector(14 - 1 downto 0);
  signal remainder_2 : std_logic_vector(14 - 1 downto 0);
  signal remainder : std_logic_vector(15 - 1 downto 0);
  signal exp_rmd : std_logic_vector(16 - 1 downto 0);
  signal product : std_logic_vector(22 - 1 downto 0);
  signal signed_part : std_logic_vector(20 - 1 downto 0);
begin
  -- premiere partie
  part_1 <= x(19 - 1 downto 14);
  -- exponentielle de cette partie
  component1 : exp_tbl_19
    port map (x => part_1,
              y => exp_tbl_out);
  exp_part1 <= ("0" & part_1) + ("00000" & exp_tbl_out);

  -- partie restante
  component2 : log_tbl_19
    port map (x => part_1,
              y => remainder_1);
  remainder_2 <= x(14 - 1 downto 0);
  remainder <= ("0" & remainder_1) + ("0" & remainder_2);
  -- exponentielle de ce reste
  component3 : exp_15
    port map (x => remainder,
              y => exp_rmd);

  -- calcul du resultat
  product <= exp_part1 * exp_rmd;
  signed_part <= (exp_part1 & "00000000000000") + ("000000" & product(22 - 1 downto 8));
  y <= ("0000" & exp_rmd) + signed_part;
end architecture;

-- exp d'un nombre avec une mantisse de 22 bits dont 21 non nuls
-- =============================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
library work;
use work.pkg_exp_tbl.all;
use work.pkg_exp.exp_19;

entity exp_21 is
  port (x : in  std_logic_vector(21 - 1 downto 0);
        y : out std_logic_vector(22 - 1 downto 0);
        sign : in std_logic);
end entity;

architecture arch of exp_21 is
  signal signed_input : std_logic_vector(5 - 1 downto 0);
  signal exp_tbl_out : std_logic_vector(20 - 1 downto 17);
  signal part_1 : std_logic_vector(21 - 1 downto 17);
  signal exp_part1 : std_logic_vector(5 - 1 downto 0);
  signal remainder_1 : std_logic_vector(18 - 1 downto 0);
  signal remainder_2 : std_logic_vector(18 - 1 downto 0);
  signal remainder : std_logic_vector(19 - 1 downto 0);
  signal exp_rmd : std_logic_vector(20 - 1 downto 0);
  signal product : std_logic_vector(25 - 1 downto 0);
  signal signed_part : std_logic_vector(22 - 1 downto 0);
begin
  -- premiere partie
  part_1 <= x(21 - 1 downto 17);
  -- exponentielle de cette partie
  signed_input(4) <= sign;
  signed_input(4 - 1 downto 0) <= part_1;
  component1 : exp_tbl_21
    port map (x => signed_input,
              y => exp_tbl_out);
  exp_part1 <= ("0" & part_1) + ("00" & exp_tbl_out) when sign = '0' else ("0" & part_1) + "00001" - ("00" & exp_tbl_out);

  -- partie restante
  component2 : log_tbl_21
    port map (x => signed_input,
              y => remainder_1);
  remainder_2 <= ("0" & x(17 - 1 downto 0)) when sign = '0' else "100000000000000000" - ("0" & x(17 - 1 downto 0));
  remainder <= ("0" & remainder_1) + ("0" & remainder_2);
  -- exponentielle de ce reste
  component3 : exp_19
    port map (x => remainder,
              y => exp_rmd);

  -- calcul du resultat
  product <= exp_part1 * exp_rmd;
  signed_part <= (exp_part1 & "00000000000000000") + ("00" & product(25 - 1 downto 5));
  y <= ("00" & exp_rmd) + signed_part when sign = '0' else ("00" & exp_rmd) - signed_part;
end architecture;

