library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

package pkg_exp is
  component exp_14 is
    port (x : in  std_logic_vector(14 - 1 downto 0);
          y : out std_logic_vector(15 - 1 downto 0));
  end component;
  component exp_19 is
    port (x : in  std_logic_vector(19 - 1 downto 0);
          y : out std_logic_vector(20 - 1 downto 0));
  end component;
  component exp_24 is
    port (x : in  std_logic_vector(24 - 1 downto 0);
          y : out std_logic_vector(25 - 1 downto 0));
  end component;
  component exp_28 is
    port (x : in  std_logic_vector(28 - 1 downto 0);
          y : out std_logic_vector(29 - 1 downto 0);
          sign : in std_logic);
  end component;
end package;

-- exp d'un nombre avec une mantisse de 29 bits dont 14 non nuls
-- =============================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
library work;
use work.pkg_exp_tbl.all;

entity exp_14 is
  port (x : in  std_logic_vector(14 - 1 downto 0);
        y : out std_logic_vector(15 - 1 downto 0));
end entity;

architecture arch of exp_14 is
  signal part_1 : std_logic_vector(14 - 1 downto 0);
  -- exponentielle du morceau moins 1
  signal exp_part1 : std_logic_vector(14 - 1 downto 0);
begin
  -- premiere partie
  part_1 <= x;
  exp_part1 <= part_1;
  y <= ("0" & exp_part1);
end architecture;

-- exp d'un nombre avec une mantisse de 29 bits dont 19 non nuls
-- =============================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
library work;
use work.pkg_exp_tbl.all;
use work.pkg_exp.exp_14;

entity exp_19 is
  port (x : in  std_logic_vector(19 - 1 downto 0);
        y : out std_logic_vector(20 - 1 downto 0));
end entity;

architecture arch of exp_19 is
  signal part_1 : std_logic_vector(19 - 1 downto 14);
  -- valeur lue dans la table pour ce morceau
  signal tbl_out : std_logic_vector(8 - 1 downto 0);
  -- exponentielle du morceau moins 1
  signal exp_part1 : std_logic_vector(19 - 1 downto 0);
  -- exponentielle des morceaux suivants moins 1
  signal exp_part2 : std_logic_vector(15 - 1 downto 0);
  -- produit des deux signaux precedents
  signal product : std_logic_vector(11 - 1 downto 0);
begin
  -- premiere partie
  part_1 <= x(19 - 1 downto 14);
  -- exponentielle de cette partie
  component1 : exp_tbl_19
    port map (x => part_1,
              y => tbl_out);
  exp_part1 <= part_1 & ("000000" & tbl_out);

  -- exponentielle de la partie restante
  component2 : exp_14
    port map (x => x(14 - 1 downto 0),
              y => exp_part2);

  -- calcul du resultat
  product <= exp_part1(19 - 1 downto 14) * exp_part2(15 - 1 downto 9);

  y <= ("0" & exp_part1) + ("00000" & exp_part2) + ("000000000000000" & product(11 - 1 downto 6));
end architecture;

-- exp d'un nombre avec une mantisse de 29 bits dont 24 non nuls
-- =============================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
library work;
use work.pkg_exp_tbl.all;
use work.pkg_exp.exp_19;

entity exp_24 is
  port (x : in  std_logic_vector(24 - 1 downto 0);
        y : out std_logic_vector(25 - 1 downto 0));
end entity;

architecture arch of exp_24 is
  signal part_1 : std_logic_vector(24 - 1 downto 18);
  signal exp_part1 : std_logic_vector(6 - 1 downto 0);
  signal remainder_1 : std_logic_vector(18 - 1 downto 0);
  signal remainder_2 : std_logic_vector(18 - 1 downto 0);
  signal remainder : std_logic_vector(19 - 1 downto 0);
  signal exp_rmd : std_logic_vector(20 - 1 downto 0);
  signal product : std_logic_vector(26 - 1 downto 0);
  signal signed_part : std_logic_vector(25 - 1 downto 0);
begin
  -- premiere partie
  part_1 <= x(24 - 1 downto 18);
  -- exponentielle de cette partie
  exp_part1 <= part_1;

  -- partie restante
  component2 : log_tbl_24
    port map (x => part_1,
              y => remainder_1);
  remainder_2 <= x(18 - 1 downto 0);
  remainder <= ("0" & remainder_1) + ("0" & remainder_2);
  -- exponentielle de ce reste
  component3 : exp_19
    port map (x => remainder,
              y => exp_rmd);

  -- calcul du resultat
  product <= exp_part1 * exp_rmd;
  signed_part <= ("0" & exp_part1 & "000000000000000000") + ("0000000000" & product(26 - 1 downto 11));
  y <= ("00000" & exp_rmd) + signed_part;
end architecture;

-- exp d'un nombre avec une mantisse de 29 bits dont 28 non nuls
-- =============================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
library work;
use work.pkg_exp_tbl.all;
use work.pkg_exp.exp_24;

entity exp_28 is
  port (x : in  std_logic_vector(28 - 1 downto 0);
        y : out std_logic_vector(29 - 1 downto 0);
        sign : in std_logic);
end entity;

architecture arch of exp_28 is
  signal signed_input : std_logic_vector(7 - 1 downto 0);
  signal exp_tbl_out : std_logic_vector(27 - 1 downto 22);
  signal part_1 : std_logic_vector(28 - 1 downto 22);
  signal exp_part1 : std_logic_vector(7 - 1 downto 0);
  signal remainder_1 : std_logic_vector(23 - 1 downto 0);
  signal remainder_2 : std_logic_vector(23 - 1 downto 0);
  signal remainder : std_logic_vector(24 - 1 downto 0);
  signal exp_rmd : std_logic_vector(25 - 1 downto 0);
  signal product : std_logic_vector(32 - 1 downto 0);
  signal signed_part : std_logic_vector(29 - 1 downto 0);
begin
  -- premiere partie
  part_1 <= x(28 - 1 downto 22);
  -- exponentielle de cette partie
  signed_input(6) <= sign;
  signed_input(6 - 1 downto 0) <= part_1;
  component1 : exp_tbl_28
    port map (x => signed_input,
              y => exp_tbl_out);
  exp_part1 <= ("0" & part_1) + ("00" & exp_tbl_out) when sign = '0' else ("0" & part_1) + "0000001" - ("00" & exp_tbl_out);

  -- partie restante
  component2 : log_tbl_28
    port map (x => signed_input,
              y => remainder_1);
  remainder_2 <= ("0" & x(22 - 1 downto 0)) when sign = '0' else "10000000000000000000000" - ("0" & x(22 - 1 downto 0));
  remainder <= ("0" & remainder_1) + ("0" & remainder_2);
  -- exponentielle de ce reste
  component3 : exp_24
    port map (x => remainder,
              y => exp_rmd);

  -- calcul du resultat
  product <= exp_part1 * exp_rmd;
  signed_part <= (exp_part1 & "0000000000000000000000") + ("0000" & product(32 - 1 downto 7));
  y <= ("0000" & exp_rmd) + signed_part when sign = '0' else ("0000" & exp_rmd) - signed_part;
end architecture;

